module Atividade06 where

import Text.Read
import Data.Foldable

data Cesta = Cesta [String] Double deriving (Eq)

data Arvore a = Folha a | Nó (Arvore a) a (Arvore a) deriving (Show, Eq)

data Expressão a =
    Constante a
  | Mul  (Expressão a) (Expressão a)
  | Soma (Expressão a) (Expressão a)
  | Sub  (Expressão a) (Expressão a)
  | Div  (Expressão a) (Expressão a)
  deriving (Show, Eq)



    -- Defina uma função arvorePossui a que retorna True caso a seja
    -- um valor presente na árvore e False caso contrário. A definição
    -- de Arvore já está dada.

arvorePossui :: Eq a => Arvore a -> a -> Bool
arvorePossui (Folha v) k  = v == k
arvorePossui (Nó e v d) k =
  (v == k) || arvorePossui e k || arvorePossui d k

    -- Defina uma função convertString2Int t que recebe uma árvore do
    -- tipo Tree [Char] e retorna uma árvore do tipo Tree (Maybe Int)
    -- em que todos os elementos são convertidos para Int quando
    -- possível.


convertString2MInt :: Arvore String -> Arvore (Maybe Int)
convertString2MInt (Folha v) = Folha $ readMaybe v
convertString2MInt (Nó e v d) =
  Nó (convertString2MInt e) (readMaybe v) (convertString2MInt d)

instance Functor Arvore where
  fmap f (Folha v) = Folha $ f v
  fmap f (Nó e v d) = Nó (fmap f e) (f v) (fmap f d)

convertString2MInt' :: Arvore String -> Arvore (Maybe Int)
convertString2MInt' = fmap readMaybe


    -- Defina uma função frutasDaArvore t que recebe uma árvore do
    -- tipo Tree [Char] e retorna o número de frutas que a árvore tem
    -- (Maybe Int). Cada nó possui uma string que informa a quantidade
    -- de frutas que aquele nó possui. Caso a string não corresponda a
    -- um número inteiro, ignorar.

frutasDaArvore :: Arvore String -> Maybe Int
frutasDaArvore a =
  Just $ frutasDaArvore' arvInt 0
  where
    arvInt = convertString2MInt a
    frutasDaArvore' (Nó e Nothing d) acc =
      frutasDaArvore' d (frutasDaArvore' e acc)
    frutasDaArvore' (Nó e (Just v) d) acc =
      v + frutasDaArvore' d (frutasDaArvore' e acc)
    frutasDaArvore' (Folha (Just v)) acc = acc + v
    frutasDaArvore' _ acc = acc

    -- Defina uma função eval expr, que recebe um argumento do tipo
    -- Expressao e retorna o resultado numérico da
    -- expressão. Expressao já está definida em Atividade06.hs.

eval :: Expressão Int -> Int
eval (Constante c) = c
eval (Mul e0 e1)   = eval e0 * eval e1
eval (Soma e0 e1)  = eval e0 + eval e1
eval (Sub e0 e1)   = eval e0 - eval e1
eval (Div e0 e1)   = eval e0 `div` eval e1

    -- Estenda Expressao adicionando o construtor Div (Expressao a)
    -- (Expressao a). Escreva uma função safeEval, que interprete a
    -- divisão como divisão inteira, mas que retorne Nothing para
    -- quando não for possível fazer a divisão.

safeEval :: Expressão Int -> Maybe Int
safeEval (Constante c) = Just c
safeEval (Mul  e0 e1)  = (*) <$> safeEval e0 <*> safeEval e1
safeEval (Soma e0 e1)  = (+) <$> safeEval e0 <*> safeEval e1
safeEval (Sub  e0 e1)  = (-) <$> safeEval e0 <*> safeEval e1
safeEval (Div  e0 e1)  = do
  v0 <- safeEval e0
  v1 <- safeEval e1
  if v1 /= 0
    then Just $ v0 `div` v1
    else Nothing

    -- Defina uma instância de Functor para valores do tipo
    -- Expressao. Deve ser possível chamar fmap sobre as
    -- expressões. Por exemplo, dada uma expressão e :: Expressao a e
    -- uma função f :: a -> b, fmap f e deve retornar uma expressão do
    -- tipo Expressao b.

instance Functor Expressão where
  fmap f (Constante c) = Constante $ f c
  fmap f (Mul e0 e1)   = Mul  (fmap f e0) (fmap f e1)
  fmap f (Soma e0 e1)  = Soma (fmap f e0) (fmap f e1)
  fmap f (Sub e0 e1)   = Sub  (fmap f e0) (fmap f e1)
  fmap f (Div e0 e1)   = Div  (fmap f e0) (fmap f e1)

    -- Uma loja possui um catálogo no qual produtos são armazenados na
    -- forma Cesta [String] Double, com a lista de string
    -- identificando os nomes dos produtos e um double representando o
    -- preço total dos produtos da cesta. Crie uma instância de Monoid
    -- para esse tipo que concatena os nomes dos produtos na cesta e
    -- some os valores desses produtos. A instância de Monoid deve
    -- cuidar dos casos em que a string é vazia, ou seja, Cesta
    -- ["notebook"] 1000 <> Cesta [] 0 == Cesta ["notebook"] 1000.

instance Semigroup Cesta where
  (Cesta xs v0) <> (Cesta yx v1) =
    Cesta (xs <> yx) (v0 + v1)

instance Monoid Cesta where
  mempty = Cesta [] 0


    -- Crie uma definição de uma instância de Show para fazer a
    -- separação com vírgulas dos produtos. Isto é, show (Cesta
    -- ["Alface","Ovo","Repolho"] 15.1) deve retornar a string
    -- "Alface,Ovo,Repolho".

instance Show Cesta where
  show (Cesta xs _) =
    foldr (\a b -> a ++ if null b then b else "," ++ b) mempty xs

    -- Faça uma função produtosCaros xs d que receba uma lista de
    -- cestas xs e um double d. Ela deve retornar uma string com os
    -- nomes dos produtos (concatenados) que custam mais do que o
    -- valor d e o valor total. Dica: utilize foldMap.

produtosCaros :: [Cesta] -> Double -> Cesta
produtosCaros xs p =
  foldMap filtraCesta xs
  where
    filtraCesta c@(Cesta _ v)
      | v <= p    = mempty
      | otherwise = c

produtosCaros2 :: [Cesta] -> Double -> Cesta
produtosCaros2 xs p =
  foldr (<>) mempty $ filter (\(Cesta _ preco) -> preco > p) xs

produtosCaros3 xs p = mconcat $ filter (\(Cesta _ preco) -> preco > p) xs


    -- Após uma revisão no catálogo da loja, o gerente percebeu que
    -- havia na base de dados produtos com preço negativo. Como isso
    -- não faz sentido, desenvolva uma função atualizaCatalogo xs que
    -- recebe uma lista de cestas da mesma forma que no exercício
    -- anterior. A função devolve uma nova lista, mas com os preços
    -- dos produtos que eram negativos agora com o sinal positivo.
atualizaCatalogo :: [Cesta] -> [Cesta]
atualizaCatalogo = map (\(Cesta es v) -> Cesta es (abs v))

    -- Crie uma função somarCarrinho que, dada uma lista de cestas,
    -- retorne o valor total daquele carrinho. Utilize fold.


pegaValor (Cesta _ v) = v

somarCarrinho :: [Cesta] -> Double
somarCarrinho xs = pegaValor $ fold xs

somarCarrinho2 :: [Cesta] -> Double
somarCarrinho2 xs = sum $ map pegaValor xs
